SuperStrict

Framework BRL.Blitz
Import BRL.Random
Import BRL.Timer
Import BRL.PNGLoader
Import BRL.GlMax2D

' default parameters
Global PAR_SEED:Int		= 42
Global PAR_SHRINK:Float	= 0.6
Global PAR_YIELD:Float		= 1.1
Global PAR_ARADIUS:Int		= 15
Global PAR_AEXP:Float		= 2
Global PAR_ALPHA:Float		= 1
Global PAR_OUTW:Int		= 512
Global PAR_OUTH:Int		= 512
Global PAR_RESX:Int		= 50
Global PAR_RESY:Int		= 50
Global PAR_MINSTR:Float	= 0.9
Global PAR_BROKENLINKS:Int	= 10
Global PAR_RUNS:Int		= 0
Global PAR_SLOWMO:Int		= 0
Global PAR_SAVE:Int		= 0

' main program

' read app arguments
For Local arg:String = EachIn appargs
	Local pos:Int = arg.Find(":")
	If pos>0
		Local key:String=arg[..pos].ToUpper()
		Local val:String=arg[pos+1..]
		' set resp. parameter
		Select key
		Case "SEED"
			PAR_SEED = Int(val)
		Case "SHRINK"
			PAR_SHRINK = Float(val)
		Case "YIELD"
			PAR_YIELD = Float(val)
		Case "RADIUS"
			PAR_ARADIUS = Int(val)
		Case "EXP"
			PAR_AEXP = Float(val)
		Case "ALPHA"
			PAR_ALPHA = Float(val)
		Case "OUTW"
			PAR_OUTW = Int(val)
		Case "OUTH"
			PAR_OUTH = Int(val)
		Case "RESX"
			PAR_RESX = Int(val)
		Case "RESY"
			PAR_RESY = Int(val)
		Case "MINSTR"
			PAR_MINSTR = Float(val)
		Case "BROKEN"
			PAR_BROKENLINKS = Int(val)
		Case "RUNS"
			PAR_RUNS = Int(val)
		End Select
	ElseIf arg.ToUpper()="SLOW"
		PAR_SLOWMO=1
	ElseIf arg.ToUpper()="SAVE"
		PAR_SAVE=1
	EndIf
Next

' create output window
SeedRnd(PAR_SEED)
SetGraphicsDriver GLMax2DDriver()
Graphics PAR_OUTW,PAR_OUTH
Local timer:TTimer = CreateTimer(50)

' generate
Local map:TAtomMap = TAtomMap.Create(PAR_RESX,PAR_RESY)
If PAR_RUNS>0 And PAR_SLOWMO=0 Then
	For Local i:Int = 0 Until PAR_RUNS
		map.Update()
	Next
EndIf

' output
Local runs:Int = 0
SetBlend(ALPHABLEND)
While Not (KeyDown(kEY_ESCAPE) Or AppTerminate())
	Cls
	If PAR_RUNS=0 Or (runs<PAR_RUNS And PAR_SLOWMO) Then
		Map.Update()
		runs :+ 1
	EndIf
	map.Draw(PAR_OUTW,PAR_OUTH)
	Flip 0
	WaitTimer(timer)
Wend
' save image?
If PAR_SAVE=1 Then
	Local file:String=RequestFile("Save Pattern...",".png - portable network graphic:png",True,CurrentDir()+"/output/")
	If file
		' draw again to backbuffer
		Cls
		map.Draw(PAR_OUTW,PAR_OUTH)
		' grab and save
		Local img:TImage=CreateImage(PAR_OUTW,PAR_OUTH,1, DYNAMICIMAGE)
		GrabImage(img,0,0)
		Local pm:TPixmap=LockImage(img)
		SavePixmapPNG(pm,file)
		UnlockImage(img)
	EndIf
EndIf
End

' app types

' atom (mesh point)
Type TAtom
	Field map:TAtomMap		' parent map
	Field x:Double,y:Double	' position
	Field fx:Double,fy:Double	' force

	' create at spec. pos
	Function Create:TAtom(pMap:TAtomMap, pX:Double, pY:Double)
		Local a:TAtom = New TAtom
		a.map = pMap
		a.x = pX
		a.y = pY
		Return a
	End Function
	' reset forces
	Method NullForces()
		fx=0
		fy=0
	End Method
	' add force vector
	Method AddForces(pX:Double, pY:Double)
		fx :+ pX
		fy :+ pY
	End Method
	' move according to effective force
	Method Move()
		x :+ fx
		y :+ fy
		' teleport if out of map border
		x = (x+map.w) Mod map.w
		y = (y+map.h) Mod map.h
	End Method
End Type

' link between to atoms
Type TAtomLink
	Field map:TAtomMap		' parent map
	Field a1:TAtom, a2:TAtom	' linked atoms
	Field ln0:Double			' initial length
	Field ln1:Double			' desired final length
	Field lne:Double			' yield point
	Field ln:Double			' actual length
	Field str:Double			' strength
	
	' create between given atoms with spec. strength
	Function Create:TAtomLink(pMap:TAtomMap, pA1:TAtom, pA2:TAtom, pStr:Double)
		Local l:TAtomLink = New TAtomLink
		l.map = pMap
		l.a1 = pA1
		l.a2 = pA2
		' calulate distance/length between atoms
		Local dx:Double, dy:Double 
		l.CalcD(dx,dy)
		l.ln0 = Sqr(dx*dx+dy*dy)
		l.ln1 = l.ln0 *PAR_SHRINK
		l.lne = l.ln0 *(pStr*PAR_YIELD +1 -pStr)
		l.str = pStr
		Return l
	End Function
	' apply link force to atoms
	Method ApplyForces()
		' calculate actual length
		Local dx:Double, dy:Double
		CalcD(dx,dy)
		ln = Sqr(dx*dx+dy*dy)
		' break link if over yield point
		If ln>lne Then str = 0
		' stretch/ compress
		Local dl:Double = ln-ln1
		' link force
		Local f:Double = dl*str*0.05
		Local fx:Double = dx*f
		Local fy:Double = dy*f
		' apply forces
		a1.AddForces(fx,fy)
		a2.AddForces(-fx,-fy)
	End Method
	' calculate distances between atoms
	Method CalcD(dx:Double Var, dy:Double Var)
		dx = a2.x -a1.x
		dy = a2.y -a1.y
		' adjust difference if going through map border
		If dx<-map.w/2 Then dx:+map.w
		If dx>map.w/2 Then dx:-map.w
		If dy<-map.h/2 Then dy:+map.h
		If dy>map.h/2 Then dy:-map.h
	End Method
End Type

' map of atoms and their links
Type TAtomMap
	Field atoms:TAtom[]		' atoms
	Field links:TAtomLink[]	' links between
	Field w:Int,h:Int			' map size
	Field img:TImage			' atom image
	
	' create with spec. size
	Function Create:TAtomMap(pW:Int, pH:Int)
		Local map:TAtomMap = New TAtomMap
		map.w=pW
		map.h=pH
		' create atoms
		map.atoms = New TAtom[pW*pH]
		For Local i:Int = 0 Until map.atoms.length
			Local x:Int = i Mod pW
			Local y:Int = i/pW
			map.atoms[i] = TAtom.Create(map,x,y)
		Next
		' set up links
		map.links = New TAtomLink[pW*pH*4]
		Local i:Int = 0
		For Local y:Int = 0 Until pH
			For Local x:Int = 0 Until pW
				map.links[i] = map.Link(x,y,x+1,y)
				map.links[i+1] = map.Link(x,y,x,y+1)
				map.links[i+2] = map.Link(x,y,x+1,y+1)
				map.links[i+3] = map.Link(x,y,x+1,y-1)
				i:+4
			Next
		Next
		' generate atom image
		map.CreateAtomImage()
		Return map
	End Function
	' link two atoms
	Method Link:TAtomLink(pX1:Int,pY1:Int, pX2:Int,PY2:Int)
		pX2 = (pX2+w) Mod w
		pY2 = (pY2+h) Mod h
		' distribute random strength
		Local str:Float = Rnd(PAR_MINSTR,1)
		' break random links
		If Rand(1000)>(1000-PAR_BROKENLINKS) Then str=0
		' create link
		Local i1:Int = pY1*w+pX1
		Local i2:Int = pY2*w+pX2
		Return TAtomLink.Create(Self, atoms[i1], atoms[i2], str)
	End Method
	' update link
	Method Update()
		' reset forces for all atoms
		For Local a:TAtom = EachIn atoms
			a.NullForces()
		Next
		' apply forces
		For Local l:TAtomLink = EachIn links
			l.ApplyForces()
		Next
		' move atoms
		For Local a:TAtom = EachIn atoms
			a.Move()
		Next
	End Method
	' draw map with spec. size in pixels
	Method Draw(pW:Float,pH:Float)
		SetColor 255,255,255
		SetAlpha PAR_ALPHA
		For Local a:TAtom = EachIn atoms
			' draw multiple times for those crossing the map border
			For Local iy:Int = -1 To 1
				For Local ix:Int = -1 To 1
					Local x:Double = a.x +ix *w
					Local y:Double = a.y +iy *h
					DrawImage(img, x/w*pW-PAR_ARADIUS, y/h*pH-PAR_ARADIUS)
				Next
			Next
		Next
	End Method
	' create image for atom drawing
	Method CreateAtomImage()
		img = CreateImage(PAR_ARADIUS*2,PAR_ARADIUS*2)
		Local pm:TPixmap = LockImage(img)
		' fill image
		For Local y:Int = 0 Until PAR_ARADIUS*2
			For Local x:Int = 0 Until PAR_ARADIUS*2
				Local a:Byte = 0		' default alpha
				Local dx:Int = x-PAR_ARADIUS
				Local dy:Int = y-PAR_ARADIUS
				Local d:Double = Sqr(dx*dx+dy*dy)
				' set alpha according to distance to image center
				If d<PAR_ARADIUS Then
					a = 255 *Cos(d/PAR_ARADIUS*90)^PAR_AEXP
				EndIf
				Local argb:Int = a Shl 24 | $ffffff
				WritePixel(pm, x,y, argb)
			Next
		Next
		UnlockImage(img)
	End Method
End Type